@@ -5,7 +5,7 @@ import fnmatch
 import os
 import shlex
 import sys
-from subprocess import Popen, PIPE
+from subprocess import run
 from typing import List, Tuple
 
 import configargparse
@@ -40,33 +40,43 @@ def main() -> None:
     args = parse_arguments()
     active_window = get_active_window()
 
-    returncode, stdout = open_main_rofi_window(args.rofi_args, read_character_files(args.files))
+    returncode, stdout = open_main_rofi_window(
+        args.rofi_args,
+        read_character_files(args.files),
+        args.prompt,
+        args.max_recent
+    )
 
     if returncode == 1:
         sys.exit()
     else:
-        characters = process_chosen_characters(stdout.splitlines(), args.skin_tone, args.rofi_args)
+        if 10 <= returncode <= 19:
+            default_handle_recent_character(returncode - 9, args, active_window)
+        else:
+            characters = process_chosen_characters(
+                stdout.splitlines(),
+                args.skin_tone,
+                args.rofi_args
+            )
+            save_characters_to_recent_file(characters, args.max_recent)
 
-        if returncode == 0:
-            if args.copy_only:
+            if returncode == 0:
+                default_handle(characters, args, active_window)
+            elif returncode == 20:
                 copy_characters_to_clipboard(characters)
-            else:
-                insert_characters(characters, active_window, args.insert_with_clipboard)
-        elif returncode == 10:
-            copy_characters_to_clipboard(characters)
-        elif returncode == 11:
-            type_characters(characters, active_window)
-        elif returncode == 12:
-            copy_paste_characters(characters, active_window)
+            elif returncode == 21:
+                type_characters(characters, active_window)
+            elif returncode == 22:
+                copy_paste_characters(characters, active_window)
 
 
 def parse_arguments() -> argparse.Namespace:
     parser = configargparse.ArgumentParser(
-        description='Select, insert or copy Unicode characters using rofi',
+        description='Select, insert or copy Unicode characters using rofi.',
         default_config_files=[os.path.join(directory, 'rofimoji.rc') for directory in
                               BaseDirectory.xdg_config_dirs]
     )
-    parser.add_argument('--version', action='version', version='rofimoji 4.1.1')
+    parser.add_argument('--version', action='version', version='rofimoji 4.2.0')
     parser.add_argument(
         '--insert-with-clipboard',
         '-p',
@@ -102,6 +112,14 @@ def parse_arguments() -> argparse.Namespace:
         metavar='FILE',
         help='Read characters from this file instead, one entry per line'
     )
+    parser.add_argument(
+        '--prompt',
+        '-r',
+        dest='prompt',
+        action='store',
+        default='😀 ',
+        help='Set rofimoj\'s  prompt'
+    )
     parser.add_argument(
         '--rofi-args',
         dest='rofi_args',
@@ -109,6 +127,14 @@ def parse_arguments() -> argparse.Namespace:
         default='',
         help='A string of arguments to give to rofi'
     )
+    parser.add_argument(
+        '--max-recent',
+        dest='max_recent',
+        action='store',
+        type=int,
+        default=10,
+        help='Show at most this number of recently used characters (cannot be larger than 10)'
+    )
 
     parsed_args = parser.parse_args()
     parsed_args.rofi_args = shlex.split(parsed_args.rofi_args)
@@ -117,17 +143,13 @@ def parse_arguments() -> argparse.Namespace:
 
 
 def get_active_window() -> str:
-    xdotool = Popen(args=['xdotool', 'getactivewindow'], stdout=PIPE)
-    return xdotool.communicate()[0].decode("utf-8")[:-1]
+    return run(args=['xdotool', 'getactivewindow'], capture_output=True, encoding='utf-8').stdout[:-1]
 
 
 def read_character_files(file_names: List[str]) -> str:
     entries = ''
 
-    if len(file_names) == 1 and file_names[0] == 'all':
-        file_names = [os.path.splitext(file)[0] for file in
-                      os.listdir(os.path.join(os.path.dirname(__file__), "data"))
-                      if fnmatch.fnmatch(file, "*.csv")]
+    file_names = resolve_all_files(file_names)
 
     for file_name in file_names:
         entries = entries + load_from_file(file_name)
@@ -135,6 +157,14 @@ def read_character_files(file_names: List[str]) -> str:
     return entries
 
 
+def resolve_all_files(file_names):
+    if len(file_names) == 1 and file_names[0] == 'all':
+        file_names = [os.path.splitext(file)[0] for file in
+                      os.listdir(os.path.join(os.path.dirname(__file__), "data"))
+                      if fnmatch.fnmatch(file, "*.csv")]
+    return file_names
+
+
 def load_from_file(file_name: str) -> str:
     provided_file = os.path.join(os.path.dirname(__file__), "data", file_name + '.csv')
     if os.path.isfile(file_name):
@@ -158,44 +188,69 @@ def load_all_characters() -> str:
     return characters
 
 
-def open_main_rofi_window(args: List[str], characters: str) -> Tuple[int, bytes]:
-    rofi = Popen(
-        [
-            'rofi',
-            '-dmenu',
-            '-markup-rows',
-            '-i',
-            '-multi-select',
-            '-p',
-            ' 😀   ',
-            '-kb-custom-1',
-            'Alt+c',
-            '-kb-custom-2',
-            'Alt+t',
-            '-kb-custom-3',
-            'Alt+p',
-            *args
-        ],
-        stdin=PIPE,
-        stdout=PIPE
+def load_recent_characters(max: int) -> List[str]:
+    try:
+        with open(os.path.join(BaseDirectory.xdg_data_home, 'rofimoji', 'recent'), 'r') as file:
+            return file.read().strip().split('\n')[:max]
+    except FileNotFoundError:
+        return []
+
+
+def format_recent_characters(max: int) -> str:
+    pairings = [f'{(index + 1) % 10}: {character}' for index, character in
+                enumerate(load_recent_characters(max))]
+
+    return ' | '.join(pairings)
+
+
+def open_main_rofi_window(rofi_args: List[str], characters: str, prompt: str, max_recent: int) -> Tuple[int, str]:
+    parameters = [
+        'rofi',
+        '-dmenu',
+        '-markup-rows',
+        '-i',
+        '-multi-select',
+        '-p',
+        prompt,
+        '-kb-custom-11',
+        'Alt+c',
+        '-kb-custom-12',
+        'Alt+t',
+        '-kb-custom-13',
+        'Alt+p',
+        *rofi_args
+    ]
+
+    recent_characters = format_recent_characters(max_recent)
+    if len(recent_characters) > 0:
+        parameters.extend(['-mesg', recent_characters])
+
+    rofi = run(
+        parameters,
+        input=characters,
+        capture_output=True,
+        encoding='utf-8'
     )
-    (stdout, _) = rofi.communicate(input=characters.encode('UTF-8'))
-    return rofi.returncode, stdout
+    return rofi.returncode, rofi.stdout
 
 
 def process_chosen_characters(
-        chosen_characters: List[bytes],
+        chosen_characters: List[str],
         skin_tone: str,
         rofi_args: List[str]
 ) -> str:
     result = ""
     for line in chosen_characters:
-        character = line.decode('utf-8').split(" ")[0]
+        character = line.split(" ")[0]
 
-        if character in skin_tone_selectable_emojis:
-            character = select_skin_tone(character, skin_tone, rofi_args)
+        characters_with_skin_tone = ''
+        for element in character:
+            if element in skin_tone_selectable_emojis:
+                characters_with_skin_tone += select_skin_tone(element, skin_tone, rofi_args)
+            else:
+                characters_with_skin_tone += element
 
-        result += character
+        result += characters_with_skin_tone
 
     return result
 
@@ -211,7 +266,7 @@ def select_skin_tone(selected_emoji: chr, skin_tone: str, rofi_args: List[str])
             fitzpatrick_modifiers.keys()
         ))
 
-        rofi_skin = Popen(
+        rofi_skin = run(
             [
                 'rofi',
                 '-dmenu',
@@ -220,41 +275,76 @@ def select_skin_tone(selected_emoji: chr, skin_tone: str, rofi_args: List[str])
                 selected_emoji + '   ',
                 *rofi_args
             ],
-            stdin=PIPE,
-            stdout=PIPE
+            input=modified_emojis,
+            capture_output=True,
+            encoding='utf-8'
         )
 
-        (stdout_skin, _) = rofi_skin.communicate(input=modified_emojis.encode('utf-8'))
-
         if rofi_skin.returncode == 1:
             return ''
 
-        return stdout_skin.split()[0].decode('utf-8')
+        return rofi_skin.stdout.split(' ')[0]
 
 
-def insert_characters(
-        characters: str,
-        active_window: str,
-        insert_with_clipboard: bool = False
-) -> None:
-    if insert_with_clipboard:
+def save_characters_to_recent_file(characters: str, max_recent_from_conf: int):
+    old_file_name = os.path.join(BaseDirectory.xdg_data_home, 'rofimoji', 'recent')
+    new_file_name = os.path.join(BaseDirectory.xdg_data_home, 'rofimoji', 'recent_temp')
+
+    max_recent = min(max_recent_from_conf, 10)
+
+    os.makedirs(os.path.dirname(new_file_name), exist_ok=True)
+    with open(new_file_name, 'w+') as new_file:
+        new_file.write(characters + '\n')
+
+        try:
+            with open(old_file_name, 'r') as old_file:
+                index = 0
+                for line in old_file:
+                    if characters == line.strip():
+                        continue
+                    if index == max_recent - 1:
+                        break
+                    new_file.write(line)
+                    index = index + 1
+
+            os.remove(old_file_name)
+        except FileNotFoundError:
+            pass
+
+    os.rename(new_file_name, old_file_name)
+
+
+def append_to_favorites_file(characters: str):
+    file_name = os.path.join(BaseDirectory.xdg_data_home, 'rofimoji', 'favorites')
+
+    os.makedirs(os.path.dirname(file_name), exist_ok=True)
+    with open(file_name, 'a+') as file:
+        file.write(characters + '\n')
+
+
+def default_handle(characters: str, args: argparse.Namespace, active_window: str):
+    if args.copy_only:
+        copy_characters_to_clipboard(characters)
+    elif args.insert_with_clipboard:
         copy_paste_characters(characters, active_window)
     else:
         type_characters(characters, active_window)
 
 
+def default_handle_recent_character(position: int, args: argparse.Namespace, active_window: str):
+    recent_characters = load_recent_characters(position)
+
+    default_handle(recent_characters[position - 1].strip(), args, active_window)
+
+
 def copy_paste_characters(characters: str, active_window: str) -> None:
-    old_clipboard_content = Popen(args=['xsel', '-o', '-b'], stdout=PIPE) \
-        .communicate()[0]
-    old_primary_content = Popen(args=['xsel', '-o', '-p'], stdout=PIPE) \
-        .communicate()[0]
+    old_clipboard_content = run(args=['xsel', '-o', '-b'], capture_output=True).stdout
+    old_primary_content = run(args=['xsel', '-o', '-p'], capture_output=True).stdout
 
-    Popen(args=['xsel', '-i', '-b'], stdin=PIPE) \
-        .communicate(input=characters.encode('utf-8'))
-    Popen(args=['xsel', '-i', '-p'], stdin=PIPE) \
-        .communicate(input=characters.encode('utf-8'))
+    run(args=['xsel', '-i', '-b'], input=characters, encoding='utf-8')
+    run(args=['xsel', '-i', '-p'], input=characters, encoding='utf-8')
 
-    Popen([
+    run([
         'xdotool',
         'windowfocus',
         '--sync',
@@ -264,19 +354,16 @@ def copy_paste_characters(characters: str, active_window: str) -> None:
         'Shift+Insert',
         'sleep',
         '0.05',
-    ]).wait()
+    ])
 
-    Popen(args=['xsel', '-i', '-b'], stdin=PIPE) \
-        .communicate(input=old_clipboard_content)
-    Popen(args=['xsel', '-i', '-p'], stdin=PIPE) \
-        .communicate(input=old_primary_content)
+    run(args=['xsel', '-i', '-b'], input=old_clipboard_content)
+    run(args=['xsel', '-i', '-p'], input=old_primary_content)
 
 
 def type_characters(characters: str, active_window: str) -> None:
-    Popen([
+    run([
         'xdotool',
         'type',
-        '--clearmodifiers',
         '--window',
         active_window,
         characters
@@ -284,15 +371,14 @@ def type_characters(characters: str, active_window: str) -> None:
 
 
 def copy_characters_to_clipboard(characters: str) -> None:
-    xsel = Popen(
-        [
-            'xsel',
-            '-i',
-            '-b'
-        ],
-        stdin=PIPE
+    run([
+        'xsel',
+        '-i',
+        '-b'
+    ],
+        input=characters,
+        encoding='utf-8'
     )
-    xsel.communicate(input=characters.encode('utf-8'))
 
 
 if __name__ == "__main__":
